@@ -33,26 +33,35 @@ import kombu.messaging
 from oslo_config import cfg
 from oslo_log import log as logging
 from oslo_utils import eventletutils
-from oslo_utils import netutils
+from oslo_utils import importutils
 import six
+import six.moves
 from six.moves.urllib import parse
 
+import oslo_messaging
 from oslo_messaging._drivers import amqp as rpc_amqp
 from oslo_messaging._drivers import amqpdriver
 from oslo_messaging._drivers import base
 from oslo_messaging._drivers import common as rpc_common
 from oslo_messaging._drivers import pool
-from oslo_messaging._i18n import _
-from oslo_messaging._i18n import _LE
-from oslo_messaging._i18n import _LI
-from oslo_messaging._i18n import _LW
 from oslo_messaging import _utils
 from oslo_messaging import exceptions
 
+eventlet = importutils.try_import('eventlet')
+if eventlet and eventletutils.is_monkey_patched("thread"):
+    # Here we initialize module with the native python threading module
+    # if it was already monkey patched by eventlet/greenlet.
+    stdlib_threading = eventlet.patcher.original('threading')
+else:
+    # Manage the case where we run this driver in a non patched environment
+    # and where user even so configure the driver to run heartbeat through
+    # a python thread, if we don't do that when the heartbeat will start
+    # we will facing an issue by trying to override the threading module.
+    stdlib_threading = threading
+
 # NOTE(sileht): don't exists in py2 socket module
 TCP_USER_TIMEOUT = 18
 
-
 rabbit_opts = [
     cfg.BoolOpt('ssl',
                 default=False,
@@ -79,6 +88,15 @@ rabbit_opts = [
                deprecated_name='kombu_ssl_ca_certs',
                help='SSL certification authority file '
                     '(valid only if SSL enabled).'),
+    cfg.BoolOpt('heartbeat_in_pthread',
+                default=False,
+                help="EXPERIMENTAL: Run the health check heartbeat thread"
+                     "through a native python thread. By default if this"
+                     "option isn't provided the  health check heartbeat will"
+                     "inherit the execution model from the parent process. By"
+                     "example if the parent process have monkey patched the"
+                     "stdlib by using eventlet/greenlet then the heartbeat"
+                     "will be run through a green thread."),
     cfg.FloatOpt('kombu_reconnect_delay',
                  default=1.0,
                  deprecated_group='DEFAULT',
@@ -101,49 +119,11 @@ rabbit_opts = [
                     'the one we are currently connected to becomes '
                     'unavailable. Takes effect only if more than one '
                     'RabbitMQ node is provided in config.'),
-    cfg.StrOpt('rabbit_host',
-               default='localhost',
-               deprecated_group='DEFAULT',
-               deprecated_for_removal=True,
-               deprecated_reason="Replaced by [DEFAULT]/transport_url",
-               help='The RabbitMQ broker address where a single node is '
-                    'used.'),
-    cfg.PortOpt('rabbit_port',
-                default=5672,
-                deprecated_group='DEFAULT',
-                deprecated_for_removal=True,
-                deprecated_reason="Replaced by [DEFAULT]/transport_url",
-                help='The RabbitMQ broker port where a single node is used.'),
-    cfg.ListOpt('rabbit_hosts',
-                default=['$rabbit_host:$rabbit_port'],
-                deprecated_group='DEFAULT',
-                deprecated_for_removal=True,
-                deprecated_reason="Replaced by [DEFAULT]/transport_url",
-                help='RabbitMQ HA cluster host:port pairs.'),
-    cfg.StrOpt('rabbit_userid',
-               default='guest',
-               deprecated_group='DEFAULT',
-               deprecated_for_removal=True,
-               deprecated_reason="Replaced by [DEFAULT]/transport_url",
-               help='The RabbitMQ userid.'),
-    cfg.StrOpt('rabbit_password',
-               default='guest',
-               deprecated_group='DEFAULT',
-               deprecated_for_removal=True,
-               deprecated_reason="Replaced by [DEFAULT]/transport_url",
-               help='The RabbitMQ password.',
-               secret=True),
     cfg.StrOpt('rabbit_login_method',
                choices=('PLAIN', 'AMQPLAIN', 'RABBIT-CR-DEMO'),
                default='AMQPLAIN',
                deprecated_group='DEFAULT',
                help='The RabbitMQ login method.'),
-    cfg.StrOpt('rabbit_virtual_host',
-               default='/',
-               deprecated_group='DEFAULT',
-               deprecated_for_removal=True,
-               deprecated_reason="Replaced by [DEFAULT]/transport_url",
-               help='The RabbitMQ virtual host.'),
     cfg.IntOpt('rabbit_retry_interval',
                default=1,
                help='How frequently to retry connecting with RabbitMQ.'),
@@ -156,12 +136,6 @@ rabbit_opts = [
                default=30,
                help='Maximum interval of RabbitMQ connection retries. '
                     'Default is 30 seconds.'),
-    cfg.IntOpt('rabbit_max_retries',
-               default=0,
-               deprecated_for_removal=True,
-               deprecated_group='DEFAULT',
-               help='Maximum number of RabbitMQ connection retries. '
-                    'Default is 0 (infinite retry count).'),
     cfg.BoolOpt('rabbit_ha_queues',
                 default=False,
                 deprecated_group='DEFAULT',
@@ -189,11 +163,17 @@ rabbit_opts = [
                default=60,
                help="Number of seconds after which the Rabbit broker is "
                "considered down if heartbeat's keep-alive fails "
-               "(0 disable the heartbeat). EXPERIMENTAL"),
+               "(0 disables heartbeat)."),
     cfg.IntOpt('heartbeat_rate',
                default=2,
                help='How often times during the heartbeat_timeout_threshold '
                'we check the heartbeat.'),
+    cfg.IntOpt('direct_mandatory_flag',
+               default=True,
+               help='Enable/Disable the RabbitMQ mandatory flag '
+               'for direct send. The direct send is used as reply,'
+               'so the MessageUndeliverable exception is raised'
+               ' in case the client queue does not exist.'),
 ]
 
 LOG = logging.getLogger(__name__)
@@ -316,11 +296,11 @@ class Consumer(object):
                         'queue': self.queue_name,
                         'err_str': exc
                         }
-                LOG.error(_LE('Internal amqp error (541) '
-                              'during queue declare,'
-                              'retrying in %(sleep_time)s seconds. '
-                              'Queue: [%(queue)s], '
-                              'error message: [%(err_str)s]'), info)
+                LOG.error('Internal amqp error (541) '
+                          'during queue declare,'
+                          'retrying in %(sleep_time)s seconds. '
+                          'Queue: [%(queue)s], '
+                          'error message: [%(err_str)s]', info)
                 time.sleep(interval)
                 self.queue.declare()
             else:
@@ -381,8 +361,7 @@ class Consumer(object):
         try:
             self.callback(RabbitMessage(message))
         except Exception:
-            LOG.exception(_LE("Failed to process message"
-                              " ... skipping it."))
+            LOG.exception("Failed to process message ... skipping it.")
             message.reject()
 
 
@@ -471,17 +450,11 @@ class Connection(object):
         # NOTE(viktors): Parse config options
         driver_conf = conf.oslo_messaging_rabbit
 
-        self.max_retries = driver_conf.rabbit_max_retries
         self.interval_start = driver_conf.rabbit_retry_interval
         self.interval_stepping = driver_conf.rabbit_retry_backoff
         self.interval_max = driver_conf.rabbit_interval_max
 
         self.login_method = driver_conf.rabbit_login_method
-        self.virtual_host = driver_conf.rabbit_virtual_host
-        self.rabbit_hosts = driver_conf.rabbit_hosts
-        self.rabbit_port = driver_conf.rabbit_port
-        self.rabbit_userid = driver_conf.rabbit_userid
-        self.rabbit_password = driver_conf.rabbit_password
         self.rabbit_ha_queues = driver_conf.rabbit_ha_queues
         self.rabbit_transient_queues_ttl = \
             driver_conf.rabbit_transient_queues_ttl
@@ -497,6 +470,35 @@ class Connection(object):
             driver_conf.kombu_missing_consumer_retry_timeout
         self.kombu_failover_strategy = driver_conf.kombu_failover_strategy
         self.kombu_compression = driver_conf.kombu_compression
+        self.heartbeat_in_pthread = driver_conf.heartbeat_in_pthread
+
+        if self.heartbeat_in_pthread:
+            # NOTE(hberaud): Experimental: threading module is in use to run
+            # the rabbitmq health check heartbeat. in some situation like
+            # with nova-api, nova need green threads to run the cells
+            # mechanismes in an async mode, so they used eventlet and
+            # greenlet to monkey patch the python stdlib and get green threads.
+            # The issue here is that nova-api run under the apache MPM prefork
+            # module and mod_wsgi. The apache prefork module doesn't support
+            # epoll and recent kernel features, and evenlet is built over epoll
+            # and libevent, so when we run the rabbitmq heartbeat we inherit
+            # from the execution model of the parent process (nova-api), and
+            # in this case we will run the heartbeat through a green thread.
+            # We want to allow users to choose between pthread and
+            # green threads if needed in some specific situations.
+            # This experimental feature allow user to use pthread in an env
+            # that doesn't support eventlet without forcing the parent process
+            # to stop to use eventlet if they need monkey patching for some
+            # specific reasons.
+            # If users want to use pthread we need to make sure that we
+            # will use the *native* threading module for
+            # initialize the heartbeat thread.
+            # Here we override globaly the previously imported
+            # threading module with the native python threading module
+            # if it was already monkey patched by eventlet/greenlet.
+            global threading
+            threading = stdlib_threading
+        self.direct_mandatory_flag = driver_conf.direct_mandatory_flag
 
         if self.ssl:
             self.ssl_version = driver_conf.ssl_version
@@ -504,52 +506,31 @@ class Connection(object):
             self.ssl_cert_file = driver_conf.ssl_cert_file
             self.ssl_ca_file = driver_conf.ssl_ca_file
 
-        # Try forever?
-        if self.max_retries <= 0:
-            self.max_retries = None
-
-        if url.virtual_host is not None:
-            virtual_host = url.virtual_host
-        else:
-            virtual_host = self.virtual_host
-
         self._url = ''
         if url.hosts:
             if url.transport.startswith('kombu+'):
-                LOG.warning(_LW('Selecting the kombu transport through the '
-                                'transport url (%s) is a experimental feature '
-                                'and this is not yet supported.'),
+                LOG.warning('Selecting the kombu transport through the '
+                            'transport url (%s) is a experimental feature '
+                            'and this is not yet supported.',
                             url.transport)
             if len(url.hosts) > 1:
                 random.shuffle(url.hosts)
-            for host in url.hosts:
-                transport = url.transport.replace('kombu+', '')
-                transport = transport.replace('rabbit', 'amqp')
-                self._url += '%s%s://%s:%s@%s:%s/%s' % (
-                    ";" if self._url else '',
-                    transport,
-                    parse.quote(host.username or ''),
-                    parse.quote(host.password or ''),
-                    self._parse_url_hostname(host.hostname) or '',
-                    str(host.port or 5672),
-                    virtual_host)
+            transformed_urls = [
+                self._transform_transport_url(url, host)
+                for host in url.hosts]
+            self._url = ';'.join(transformed_urls)
         elif url.transport.startswith('kombu+'):
             # NOTE(sileht): url have a + but no hosts
             # (like kombu+memory:///), pass it to kombu as-is
             transport = url.transport.replace('kombu+', '')
-            self._url = "%s://%s" % (transport, virtual_host)
-        else:
-            if len(self.rabbit_hosts) > 1:
-                random.shuffle(self.rabbit_hosts)
-            for adr in self.rabbit_hosts:
-                hostname, port = netutils.parse_host_port(
-                    adr, default_port=self.rabbit_port)
-                self._url += '%samqp://%s:%s@%s:%s/%s' % (
-                    ";" if self._url else '',
-                    parse.quote(self.rabbit_userid, ''),
-                    parse.quote(self.rabbit_password, ''),
-                    self._parse_url_hostname(hostname), port,
-                    virtual_host)
+            self._url = "%s://" % transport
+            if url.virtual_host:
+                self._url += url.virtual_host
+        elif not url.hosts:
+            host = oslo_messaging.transport.TransportHost('')
+            self._url = self._transform_transport_url(
+                url, host, default_username='guest', default_password='guest',
+                default_hostname='localhost')
 
         self._initial_pid = os.getpid()
 
@@ -675,7 +656,19 @@ class Connection(object):
         try:
             return cls._SSL_PROTOCOLS[key]
         except KeyError:
-            raise RuntimeError(_("Invalid SSL version : %s") % version)
+            raise RuntimeError("Invalid SSL version : %s" % version)
+
+    def _transform_transport_url(self, url, host, default_username='',
+                                 default_password='', default_hostname=''):
+        transport = url.transport.replace('kombu+', '')
+        transport = transport.replace('rabbit', 'amqp')
+        return '%s://%s:%s@%s:%s/%s' % (
+            transport,
+            parse.quote(host.username or default_username),
+            parse.quote(host.password or default_password),
+            self._parse_url_hostname(host.hostname) or default_hostname,
+            str(host.port or 5672),
+            url.virtual_host or '')
 
     def _parse_url_hostname(self, hostname):
         """Handles hostname returned from urlparse and checks whether it's
@@ -708,25 +701,29 @@ class Connection(object):
 
     @staticmethod
     def _on_connection_blocked(reason):
-        LOG.error(_LE("The broker has blocked the connection: %s"), reason)
+        LOG.error("The broker has blocked the connection: %s", reason)
 
     @staticmethod
     def _on_connection_unblocked():
-        LOG.info(_LI("The broker has unblocked the connection"))
+        LOG.info("The broker has unblocked the connection")
 
     def ensure_connection(self):
         # NOTE(sileht): we reset the channel and ensure
         # the kombu underlying connection works
+        def on_error(exc, interval):
+            LOG.error("Connection failed: %s (retrying in %s seconds)",
+                      str(exc), interval)
+
         self._set_current_channel(None)
-        self.ensure(method=self.connection.connect)
+        self.connection.ensure_connection(errback=on_error)
+        self._set_current_channel(self.connection.channel())
         self.set_transport_socket_timeout()
 
     def ensure(self, method, retry=None,
                recoverable_error_callback=None, error_callback=None,
                timeout_is_error=True):
         """Will retry up to retry number of times.
-        retry = None means use the value of rabbit_max_retries
-        retry = -1 means to retry forever
+        retry = None or -1 means to retry forever
         retry = 0 means no retry
         retry = N means N retries
 
@@ -735,16 +732,14 @@ class Connection(object):
 
         current_pid = os.getpid()
         if self._initial_pid != current_pid:
-            LOG.warning(_LW("Process forked after connection established! "
-                            "This can result in unpredictable behavior. "
-                            "See: https://docs.openstack.org/oslo.messaging/"
-                            "latest/reference/transport.html"))
+            LOG.warning("Process forked after connection established! "
+                        "This can result in unpredictable behavior. "
+                        "See: https://docs.openstack.org/oslo.messaging/"
+                        "latest/reference/transport.html")
             self._initial_pid = current_pid
 
-        if retry is None:
-            retry = self.max_retries
         if retry is None or retry < 0:
-            retry = None
+            retry = float('inf')
 
         def on_error(exc, interval):
             LOG.debug("[%s] Received recoverable error from kombu:"
@@ -761,15 +756,15 @@ class Connection(object):
             info.update(self._get_connection_info(conn_error=True))
 
             if 'Socket closed' in six.text_type(exc):
-                LOG.error(_LE('[%(connection_id)s] AMQP server'
-                              ' %(hostname)s:%(port)s closed'
-                              ' the connection. Check login credentials:'
-                              ' %(err_str)s'), info)
+                LOG.error('[%(connection_id)s] AMQP server'
+                          ' %(hostname)s:%(port)s closed'
+                          ' the connection. Check login credentials:'
+                          ' %(err_str)s', info)
             else:
-                LOG.error(_LE('[%(connection_id)s] AMQP server on '
-                              '%(hostname)s:%(port)s is unreachable: '
-                              '%(err_str)s. Trying again in '
-                              '%(sleep_time)d seconds.'), info)
+                LOG.error('[%(connection_id)s] AMQP server on '
+                          '%(hostname)s:%(port)s is unreachable: '
+                          '%(err_str)s. Trying again in '
+                          '%(sleep_time)d seconds.', info)
 
             # XXX(nic): when reconnecting to a RabbitMQ cluster
             # with mirrored queues in use, the attempt to release the
@@ -793,9 +788,9 @@ class Connection(object):
             self._set_current_channel(new_channel)
             self.set_transport_socket_timeout()
 
-            LOG.info(_LI('[%(connection_id)s] Reconnected to AMQP server on '
-                         '%(hostname)s:%(port)s via [%(transport)s] client '
-                         'with port %(client_port)s.'),
+            LOG.info('[%(connection_id)s] Reconnected to AMQP server on '
+                     '%(hostname)s:%(port)s via [%(transport)s] client '
+                     'with port %(client_port)s.',
                      self._get_connection_info())
 
         def execute_method(channel):
@@ -818,6 +813,10 @@ class Connection(object):
             # NOTE(sileht): we must reraise this without
             # trigger error_callback
             raise
+        except exceptions.MessageUndeliverable:
+            # NOTE(gsantomaggio): we must reraise this without
+            # trigger error_callback
+            raise
         except Exception as exc:
             error_callback and error_callback(exc)
             self._set_current_channel(None)
@@ -825,12 +824,17 @@ class Connection(object):
             # is still broken
             info = {'err_str': exc, 'retry': retry}
             info.update(self.connection.info())
-            msg = _('Unable to connect to AMQP server on '
-                    '%(hostname)s:%(port)s after %(retry)s '
-                    'tries: %(err_str)s') % info
+            msg = ('Unable to connect to AMQP server on '
+                   '%(hostname)s:%(port)s after %(retry)s '
+                   'tries: %(err_str)s' % info)
             LOG.error(msg)
             raise exceptions.MessageDeliveryFailure(msg)
 
+    @staticmethod
+    def on_return(exception, exchange, routing_key, message):
+        raise exceptions.MessageUndeliverable(exception, exchange, routing_key,
+                                              message)
+
     def _set_current_channel(self, new_channel):
         """Change the channel to use.
 
@@ -849,7 +853,8 @@ class Connection(object):
         if new_channel is not None:
             if self.purpose == rpc_common.PURPOSE_LISTEN:
                 self._set_qos(new_channel)
-            self._producer = kombu.messaging.Producer(new_channel)
+            self._producer = kombu.messaging.Producer(new_channel,
+                                                      on_return=self.on_return)
             for consumer in self._consumers:
                 consumer.declare(self)
 
@@ -864,11 +869,11 @@ class Connection(object):
         """Close/release this connection."""
         self._heartbeat_stop()
         if self.connection:
-            for consumer, tag in self._consumers.items():
-                if consumer.type == 'fanout':
-                    LOG.debug('[connection close] Deleting fanout '
-                              'queue: %s ' % consumer.queue.name)
-                    consumer.queue.delete()
+            for consumer in six.moves.filter(lambda c: c.type == 'fanout',
+                                             self._consumers):
+                LOG.debug('[connection close] Deleting fanout '
+                          'queue: %s ' % consumer.queue.name)
+                consumer.queue.delete()
             self._set_current_channel(None)
             self.connection.release()
             self.connection = None
@@ -893,8 +898,8 @@ class Connection(object):
         if self.connection.supports_heartbeats:
             return True
         elif not self._heartbeat_support_log_emitted:
-            LOG.warning(_LW("Heartbeat support requested but it is not "
-                            "supported by the kombu driver or the broker"))
+            LOG.warning("Heartbeat support requested but it is not "
+                        "supported by the kombu driver or the broker")
             self._heartbeat_support_log_emitted = True
         return False
 
@@ -957,9 +962,9 @@ class Connection(object):
 
     def _heartbeat_start(self):
         if self._heartbeat_supported_and_enabled():
-            self._heartbeat_exit_event = eventletutils.Event()
+            self._heartbeat_exit_event = threading.Event()
             self._heartbeat_thread = threading.Thread(
-                target=self._heartbeat_thread_job)
+                target=self._heartbeat_thread_job, name="Rabbit-heartbeat")
             self._heartbeat_thread.daemon = True
             self._heartbeat_thread.start()
         else:
@@ -1010,12 +1015,12 @@ class Connection(object):
                             ConnectRefuseError,
                             OSError,
                             kombu.exceptions.OperationalError) as exc:
-                        LOG.info(_LI("A recoverable connection/channel error "
-                                     "occurred, trying to reconnect: %s"), exc)
+                        LOG.info("A recoverable connection/channel error "
+                                 "occurred, trying to reconnect: %s", exc)
                         self.ensure_connection()
                 except Exception:
-                    LOG.warning(_LW("Unexpected error during heartbeat "
-                                "thread processing, retrying..."))
+                    LOG.warning("Unexpected error during heartbeat "
+                                "thread processing, retrying...")
                     LOG.debug('Exception', exc_info=True)
 
             self._heartbeat_exit_event.wait(
@@ -1029,8 +1034,8 @@ class Connection(object):
 
         def _connect_error(exc):
             log_info = {'topic': consumer.routing_key, 'err_str': exc}
-            LOG.error(_LE("Failed to declare consumer for topic '%(topic)s': "
-                          "%(err_str)s"), log_info)
+            LOG.error("Failed to declare consumer for topic '%(topic)s': "
+                      "%(err_str)s", log_info)
 
         def _declare_consumer():
             consumer.declare(self)
@@ -1063,8 +1068,7 @@ class Connection(object):
 
         def _error_callback(exc):
             _recoverable_error_callback(exc)
-            LOG.error(_LE('Failed to consume message from queue: %s'),
-                      exc)
+            LOG.error('Failed to consume message from queue: %s', exc)
 
         def _consume():
             # NOTE(sileht): in case the acknowledgment or requeue of a
@@ -1116,12 +1120,12 @@ class Connection(object):
         responses for call/multicall
         """
 
-        consumer = Consumer(exchange_name=topic,
+        consumer = Consumer(exchange_name='',  # using default exchange
                             queue_name=topic,
-                            routing_key=topic,
+                            routing_key='',
                             type='direct',
                             durable=False,
-                            exchange_auto_delete=True,
+                            exchange_auto_delete=False,
                             queue_auto_delete=False,
                             callback=callback,
                             rabbit_ha_queues=self.rabbit_ha_queues,
@@ -1165,16 +1169,17 @@ class Connection(object):
         self.declare_consumer(consumer)
 
     def _ensure_publishing(self, method, exchange, msg, routing_key=None,
-                           timeout=None, retry=None):
+                           timeout=None, retry=None, transport_options=None):
         """Send to a publisher based on the publisher class."""
 
         def _error_callback(exc):
             log_info = {'topic': exchange.name, 'err_str': exc}
-            LOG.error(_LE("Failed to publish message to topic "
-                          "'%(topic)s': %(err_str)s"), log_info)
+            LOG.error("Failed to publish message to topic "
+                      "'%(topic)s': %(err_str)s", log_info)
             LOG.debug('Exception', exc_info=exc)
 
-        method = functools.partial(method, exchange, msg, routing_key, timeout)
+        method = functools.partial(method, exchange, msg, routing_key,
+                                   timeout, transport_options)
 
         with self._connection_lock:
             self.ensure(method, retry=retry, error_callback=_error_callback)
@@ -1188,15 +1193,16 @@ class Connection(object):
         info = dict([(k, ci.get(k)) for k in
                      ['hostname', 'port', 'transport']])
         client_port = None
-        if (not conn_error and self.channel
-                and hasattr(self.channel.connection, 'sock')
-                and self.channel.connection.sock):
+        if (not conn_error and self.channel and
+                hasattr(self.channel.connection, 'sock') and
+                self.channel.connection.sock):
             client_port = self.channel.connection.sock.getsockname()[1]
         info.update({'client_port': client_port,
                      'connection_id': self.connection_id})
         return info
 
-    def _publish(self, exchange, msg, routing_key=None, timeout=None):
+    def _publish(self, exchange, msg, routing_key=None, timeout=None,
+                 transport_options=None):
         """Publish a message."""
 
         if not (exchange.passive or exchange.name in self._declared_exchanges):
@@ -1205,21 +1211,25 @@ class Connection(object):
 
         log_info = {'msg': msg,
                     'who': exchange or 'default',
-                    'key': routing_key}
+                    'key': routing_key,
+                    'transport_options': str(transport_options)}
         LOG.trace('Connection._publish: sending message %(msg)s to'
                   ' %(who)s with routing key %(key)s', log_info)
-
         # NOTE(sileht): no need to wait more, caller expects
         # a answer before timeout is reached
         with self._transport_socket_timeout(timeout):
-            self._producer.publish(msg,
-                                   exchange=exchange,
-                                   routing_key=routing_key,
-                                   expiration=timeout,
-                                   compression=self.kombu_compression)
+            self._producer.publish(
+                msg,
+                mandatory=transport_options.at_least_once if
+                transport_options else False,
+                exchange=exchange,
+                routing_key=routing_key,
+                expiration=timeout,
+                compression=self.kombu_compression)
 
     def _publish_and_creates_default_queue(self, exchange, msg,
-                                           routing_key=None, timeout=None):
+                                           routing_key=None, timeout=None,
+                                           transport_options=None):
         """Publisher that declares a default queue
 
         When the exchange is missing instead of silently creates an exchange
@@ -1256,7 +1266,8 @@ class Connection(object):
 
     def _publish_and_raises_on_missing_exchange(self, exchange, msg,
                                                 routing_key=None,
-                                                timeout=None):
+                                                timeout=None,
+                                                transport_options=None):
         """Publisher that raises exception if exchange is missing."""
         if not exchange.passive:
             raise RuntimeError("_publish_and_retry_on_missing_exchange() must "
@@ -1264,7 +1275,7 @@ class Connection(object):
 
         try:
             self._publish(exchange, msg, routing_key=routing_key,
-                          timeout=timeout)
+                          timeout=timeout, transport_options=transport_options)
             return
         except self.connection.channel_errors as exc:
             if exc.code == 404:
@@ -1282,16 +1293,19 @@ class Connection(object):
 
     def direct_send(self, msg_id, msg):
         """Send a 'direct' message."""
-        exchange = kombu.entity.Exchange(name=msg_id,
+        exchange = kombu.entity.Exchange(name='',  # using default exchange
                                          type='direct',
                                          durable=False,
                                          auto_delete=True,
                                          passive=True)
-
+        options = oslo_messaging.TransportOptions(
+            at_least_once=self.direct_mandatory_flag)
         self._ensure_publishing(self._publish_and_raises_on_missing_exchange,
-                                exchange, msg, routing_key=msg_id)
+                                exchange, msg, routing_key=msg_id,
+                                transport_options=options)
 
-    def topic_send(self, exchange_name, topic, msg, timeout=None, retry=None):
+    def topic_send(self, exchange_name, topic, msg, timeout=None, retry=None,
+                   transport_options=None):
         """Send a 'topic' message."""
         exchange = kombu.entity.Exchange(
             name=exchange_name,
@@ -1301,7 +1315,8 @@ class Connection(object):
 
         self._ensure_publishing(self._publish, exchange, msg,
                                 routing_key=topic, timeout=timeout,
-                                retry=retry)
+                                retry=retry,
+                                transport_options=transport_options)
 
     def fanout_send(self, topic, msg, retry=None):
         """Send a 'fanout' message."""
