@@ -89,6 +89,11 @@ class TestHeartbeat(test_utils.BaseTestCase):
             info='A recoverable connection/channel error occurred, '
             'trying to reconnect: %s')
 
+    def test_run_heartbeat_in_pthread(self):
+        self.config(heartbeat_in_pthread=True,
+                    group="oslo_messaging_rabbit")
+        self._do_test_heartbeat_sent()
+
 
 class TestRabbitQos(test_utils.BaseTestCase):
 
@@ -123,7 +128,8 @@ class TestRabbitDriverLoad(test_utils.BaseTestCase):
         ('rabbit+memory', dict(transport_url='kombu+memory:/'))
     ]
 
-    @mock.patch('oslo_messaging._drivers.impl_rabbit.Connection.ensure')
+    @mock.patch('oslo_messaging._drivers.impl_rabbit.Connection'
+                '.ensure_connection')
     @mock.patch('oslo_messaging._drivers.impl_rabbit.Connection.reset')
     def test_driver_load(self, fake_ensure, fake_reset):
         self.config(heartbeat_timeout_threshold=60,
@@ -155,7 +161,8 @@ class TestRabbitDriverLoadSSL(test_utils.BaseTestCase):
                                                 cert_reqs=ssl.CERT_REQUIRED))),
     ]
 
-    @mock.patch('oslo_messaging._drivers.impl_rabbit.Connection.ensure')
+    @mock.patch('oslo_messaging._drivers.impl_rabbit.Connection'
+                '.ensure_connection')
     @mock.patch('kombu.connection.Connection')
     def test_driver_load(self, connection_klass, fake_ensure):
         self.config(group="oslo_messaging_rabbit", **self.options)
@@ -197,6 +204,7 @@ class TestRabbitPublisher(test_utils.BaseTestCase):
             'msg', expiration=1,
             exchange=exchange_mock,
             compression=self.conf.oslo_messaging_rabbit.kombu_compression,
+            mandatory=False,
             routing_key='routing_key')
 
     @mock.patch('kombu.messaging.Producer.publish')
@@ -210,6 +218,7 @@ class TestRabbitPublisher(test_utils.BaseTestCase):
             conn._publish(exchange_mock, 'msg', routing_key='routing_key')
         fake_publish.assert_called_with(
             'msg', expiration=None,
+            mandatory=False,
             compression=self.conf.oslo_messaging_rabbit.kombu_compression,
             exchange=exchange_mock,
             routing_key='routing_key')
@@ -398,6 +407,9 @@ class TestRabbitTransportURL(test_utils.BaseTestCase):
         ('rabbit_ipv4',
          dict(url='rabbit://user:password@10.20.30.40:10/vhost',
               expected=['amqp://user:password@10.20.30.40:10/vhost'])),
+        ('rabbit_no_vhost_slash',
+         dict(url='rabbit://user:password@10.20.30.40:10',
+              expected=['amqp://user:password@10.20.30.40:10/'])),
     ]
 
     def setUp(self):
@@ -406,7 +418,8 @@ class TestRabbitTransportURL(test_utils.BaseTestCase):
         self.config(heartbeat_timeout_threshold=0,
                     group='oslo_messaging_rabbit')
 
-    @mock.patch('oslo_messaging._drivers.impl_rabbit.Connection.ensure')
+    @mock.patch('oslo_messaging._drivers.impl_rabbit.Connection'
+                '.ensure_connection')
     @mock.patch('oslo_messaging._drivers.impl_rabbit.Connection.reset')
     def test_transport_url(self, fake_reset, fake_ensure):
         transport = oslo_messaging.get_transport(self.conf, self.url)
@@ -946,9 +959,9 @@ class RpcKombuHATestCase(test_utils.BaseTestCase):
 
     def setUp(self):
         super(RpcKombuHATestCase, self).setUp()
-        self.brokers = ['host1', 'host2', 'host3', 'host4', 'host5']
-        self.config(rabbit_hosts=self.brokers,
-                    rabbit_retry_interval=0.01,
+        transport_url = 'rabbit:/host1,host2,host3,host4,host5/'
+        self.messaging_conf.transport_url = transport_url
+        self.config(rabbit_retry_interval=0.01,
                     rabbit_retry_backoff=0.01,
                     kombu_reconnect_delay=0,
                     heartbeat_timeout_threshold=0,
@@ -991,6 +1004,7 @@ class RpcKombuHATestCase(test_utils.BaseTestCase):
 
 class ConnectionLockTestCase(test_utils.BaseTestCase):
     def _thread(self, lock, sleep, heartbeat=False):
+
         def thread_task():
             if heartbeat:
                 with lock.for_heartbeat():
